# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""All illustris utilities for loading simulation data.

Adapted from: https://github.com/illustristng/illustris_python
"""

import skais_mapper.illustris.util
import skais_mapper.illustris.snapshots
import skais_mapper.illustris.groupcat #noqa
